/*
    padGnneco is a pad plugin for PS emulator.
    Copyright (c) 2003 yokota (shirouto_yokota-lj@infoseek.jp)

    padGnneco.c is a derivative of padXwin v1.6

 * Pad for Psemu Pro like Emulators
 *
 * By: linuzappz <linuzappz@hotmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "padGnneco.h"

#if defined (__WIN32__)
#include <windows.h>
#include "pgn_res.h"
static char * LibName = LIB_NAME " (" PGN_LANGSTR ", DI" PGN_STRINGIFY(DI_VER) ")";

#endif

const unsigned char version = 1;	// PSEmu 1.x library
const unsigned char revision = VERSION;
const unsigned char build = BUILD;

static int pgn_init = 0;
static int pgn_pad_opened = 0;
static int pgn_pad_flags = 0;
static int ps2e = 0;

char * CALLBACK PSEgetLibName(void)
{
	return LibName;
}

unsigned long CALLBACK PSEgetLibType(void)
{
	return 8;
}

unsigned long CALLBACK PSEgetLibVersion(void)
{
	return version << 16 | revision << 8 | build;
}

long CALLBACK PADinit(long flags)
{
	pgn_pad_flags |= flags;
	if (pgn_init) return 0;
	pgn_load_conf();
	pgn_init = 1;
	return 0;
}

long CALLBACK PADshutdown(void)
{
	pgn_init = 0;
	pgn_pad_flags = 0;
	return 0;
}

long CALLBACK PADopen(void *Disp)
{
	int i = 0;
	int id;

	for (i = 0, id = 0; id < PGN_PADS; id++) {
		if ((1 << id) & pgn_pad_flags) {
			if (i == pgn_pad_opened) {
				break;
			}
			i++;
		}
	}
	pgn_open(Disp, id, ps2e);
	pgn_pad_opened++;
	return 0;
}

long CALLBACK PADclose(void)
{
	pgn_close();
	pgn_pad_opened = 0;
	return 0;
}

long CALLBACK PADquery(void)
{
	long data = 0;
	int i;

	for (i = 0; i < PGN_PADS; i++) {
		data |= 1 << i;
	}
	return data;
}

typedef struct {
	unsigned char controllerType;
	unsigned short buttonStatus;
	unsigned char rightJoyX, rightJoyY, leftJoyX, leftJoyY;
	unsigned char moveX, moveY;
	unsigned char negConWheel, negConB1, negConB2, negConB3;
	unsigned char reserved[87];
} PadDataS;

long CALLBACK PADreadPort1(PadDataS * pad)
{
	unsigned char buf[6];
	unsigned char id;

	id = pgn_read_port(buf, 0);
	pad->controllerType = (id >> 4) & 0x0f;
	pad->buttonStatus = (unsigned short) buf[1] << 8 | buf[0];
	if (id == 0x23) {
		pad->rightJoyX = buf[2];
		pad->rightJoyY = buf[3];
		pad->leftJoyX = buf[4];
		pad->leftJoyY = buf[5];

		pad->negConWheel = buf[2];
		pad->negConB1 = buf[3];
		pad->negConB2 = buf[4];
		pad->negConB3 = buf[5];
	}
	return 0;
}

long CALLBACK PADreadPort2(PadDataS * pad)
{
	unsigned char buf[6];
	unsigned char id;

	id = pgn_read_port(buf, 1);
	pad->controllerType = (id >> 4) & 0x0f;
	pad->buttonStatus = (unsigned short) buf[1] << 8 | buf[0];
	if (id == 0x23) {
		pad->rightJoyX = buf[2];
		pad->rightJoyY = buf[3];
		pad->leftJoyX = buf[4];
		pad->leftJoyY = buf[5];

		pad->negConWheel = buf[2];
		pad->negConB1 = buf[3];
		pad->negConB2 = buf[4];
		pad->negConB3 = buf[5];
	}
	return 0;
}

long CALLBACK PADconfigure(void)
{
	if (!pgn_init) {
		pgn_load_conf();
	}
	pgn_configure();
	return 0;
}

void CALLBACK PADabout(void)
{
	pgn_about();
	return;
}

long CALLBACK PADtest(void)
{
	return 0;
}

static unsigned char pad_buf[10] = { 0x00, 0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff, 0xff };
static int pad_buf_pos = 0;

unsigned char CALLBACK PADstartPoll(int num)
{
	unsigned char buf[6];
	int i;

	pad_buf_pos = 0;
	pad_buf[1] = pgn_read_port(buf, num - 1);
	pad_buf[2] = 0x5a;
	for (i = 0; i < (pad_buf[1] & 0x0f) * 2; i++) {
		pad_buf[i + 3] = buf[i];
	}

	return pad_buf[pad_buf_pos++];
}

unsigned char CALLBACK PADpoll(unsigned char cmd)
{
	if (pad_buf_pos > 9) return 0xff;
	return pad_buf[pad_buf_pos++];
}

/* ps2e */
typedef struct {
	unsigned key;
	unsigned event;
} keyEvent;
keyEvent * CALLBACK PADkeyEvent()
{
	return NULL;
}
char * CALLBACK PS2EgetLibName(void)
{
	ps2e = 1;
	return LibName;
}
unsigned long CALLBACK PS2EgetLibType(void)
{
	ps2e = 1;
	return 2;
}
unsigned long CALLBACK PS2EgetLibVersion(void)
{
	ps2e = 1;
	return version << 16 | revision << 8 | build;
}
unsigned int CALLBACK PS2EgetLibVersion2(unsigned int type)
{
	ps2e = 1;
	return 0x0002 << 16 | revision << 8 | build;
}

