/*
    padGnneco is a pad plugin for PS emulator.
    Copyright (c) 2003 yokota (shirouto_yokota-lj@infoseek.jp)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#if DI_VER == 7
#define DIRECTINPUT_VERSION 0x0700
#define PGN_LPDIRECTINPUT LPDIRECTINPUT7
#define PGN_LPDIRECTINPUTDEVICE LPDIRECTINPUTDEVICE7
#elif DI_VER == 8
#define DIRECTINPUT_VERSION 0x0800
#define PGN_LPDIRECTINPUT LPDIRECTINPUT8
#define PGN_LPDIRECTINPUTDEVICE LPDIRECTINPUTDEVICE8
#else 
#define DIRECTINPUT_VERSION 0x0500
#define PGN_LPDIRECTINPUT LPDIRECTINPUT
#define PGN_LPDIRECTINPUTDEVICE LPDIRECTINPUTDEVICE2
#endif

#include <windows.h>
#include <dinput.h>
#include <stdio.h>
#include "pgn_res.h"
#include "padGnneco.h"

#define PGN_MAX_DEV 6
#define DELAY1 50
#define DELAY2 250
#define WAIT1 5000
#define WAIT2 5000
#define PGN_MAX_CANDIDATES 6

static HINSTANCE hinst;
static PGN_LPDIRECTINPUT dinput = NULL;
static PGN_LPDIRECTINPUTDEVICE di_dev[PGN_MAX_DEV];
static GUID guid[PGN_MAX_DEV];
static int threshold;
static int threshold_list[] = { 1000, 10, 100, 10000, 100000, 1000000, 0 };
static int af_timer[PGN_PADS][NUM_DLG_AF];
	static int pushed = 0;
typedef struct {
	char k_state[256];
	DIJOYSTATE j_state[PGN_MAX_DEV - 1];
}inputstates;
static inputstates input_states_conf[2];
static inputstates input_state;

typedef struct {
	int device;
	int key;
	int analog;
	int max;
	int min;
}keyconfig;

/*
 * { down, up, right, left, a, b, r, start,
 *   neji-right, neji-left, analog-1, analog-2, analog-l,
 *   select, l2, r2,
 *   <autofire> down, up, right, left, a, b, r, start,
 *   analog-1, analog-2, analog-l, select, l2, r2 </autofire> };
 */
static keyconfig key_config_tmp[PGN_PADS][NUM_KEY_CFGS];
static keyconfig key_config[PGN_PADS][NUM_KEY_CFGS];
static keyconfig key_config_candidate[PGN_MAX_CANDIDATES];
static const keyconfig key_config_default[2][NUM_KEY_CFGS] = {
	{
		{ 2, DIJOFS_Y, 0, 65535, 49151 },
		{ 2, DIJOFS_Y, 0, 0, 16384 },
		{ 2, DIJOFS_X, 0, 65535, 49151 },
		{ 2, DIJOFS_X, 0, 0, 16384 },
		{ 2, DIJOFS_BUTTON0 , 0, 0, 0 },
		{ 2, DIJOFS_BUTTON1 , 0, 0, 0 },
		{ 2, DIJOFS_BUTTON2 , 0, 0, 0 },
		{ 2, DIJOFS_BUTTON3 , 0, 0, 0 },
		{ 2, DIJOFS_SLIDER(1), 127, 65535, 32768 },
		{ 2, DIJOFS_SLIDER(1), 128, 0, 32768 },
		{ 2, DIJOFS_SLIDER(0), 255, 65535, 0 },
		{ 2, DIJOFS_Z, 255, 65535, 0 },
		{ 2, DIJOFS_RZ, 255, 65535, 0 },

		{ 2, DIJOFS_BUTTON4, 0, 0, 0 },
		{ 2, DIJOFS_BUTTON5, 0, 0, 0 },
		{ 2, DIJOFS_BUTTON6, 0, 0, 0 },

		{ 2, DIJOFS_Y, -1, 65535, 49151 },
		{ 2, DIJOFS_Y, -1, 0, 16384 },
		{ 2, DIJOFS_X, -1, 65535, 49151 },
		{ 2, DIJOFS_X, -1, 0, 16384 },
		{ 2, DIJOFS_BUTTON0 , -1, 0, 0 },
		{ 2, DIJOFS_BUTTON1 , -1, 0, 0 },
		{ 2, DIJOFS_BUTTON2 , -1, 0, 0 },
		{ 2, DIJOFS_BUTTON3 , -1, 0, 0 },
		{ 2, DIJOFS_SLIDER(0), -1, 65535, 32768 },
		{ 2, DIJOFS_Z, -1, 65535, 32768 },
		{ 2, DIJOFS_RZ, -1, 65535, 32768 },
		{ 2, DIJOFS_BUTTON4, -1, 0, 0 },
		{ 2, DIJOFS_BUTTON5, -1, 0, 0 },
		{ 2, DIJOFS_BUTTON6, -1, 0, 0 },

		{ 2, DIJOFS_SLIDER(0), 1 | 0<<4 | 4<<8 | (160>>2)<<12 | (156>>2)<<18 | (40>>2)<<24 | 1<<30, 65535, 0 },
		{ 2, DIJOFS_Z, 3 | 2<<4 | 6<<8 | (160>>2)<<12 | (156>>2)<<18 | (40>>2)<<24 | 1<<30, 65535, 0 },

		{ 2, DIJOFS_BUTTON0 , 0x6800, 0, 0 },
		{ 2, DIJOFS_BUTTON1 , 0x9400, 0, 0 },
		{ 2, DIJOFS_BUTTON2 , 0, 0, 0 },
		{ 2, DIJOFS_RZ, 0, 65535, 32768 },

		{ 2, DIJOFS_BUTTON7, 1, 0, 0 },
	}, {
		{ 1, DIK_DOWN, 0, 0, 0 },
		{ 1, DIK_UP, 0, 0, 0 },
		{ 1, DIK_RIGHT, 0, 0, 0 },
		{ 1, DIK_LEFT, 0, 0, 0 },
		{ 1, DIK_D , 0, 0, 0 },
		{ 1, DIK_S , 0, 0, 0 },
		{ 1, DIK_F , 0, 0, 0 },
		{ 1, DIK_RETURN , 0, 0, 0 },
		{ 1, DIK_PERIOD, 127, 0, 0 },
		{ 1, DIK_COMMA, 128, 0, 0 },
		{ 1, DIK_X, 255, 0, 0 },
		{ 1, DIK_Z, 255, 0, 0 },
		{ 1, DIK_A, 255, 0, 0 },

		{ 1, DIK_SPACE, 0, 0, 0 },
		{ 1, DIK_Q, 0, 0, 0 },
		{ 1, DIK_R, 0, 0, 0 },

		{ 1, DIK_DOWN, -1, 0, 0 },
		{ 1, DIK_UP, -1, 0, 0 },
		{ 1, DIK_RIGHT, -1, 0, 0 },
		{ 1, DIK_LEFT, -1, 0, 0 },
		{ 1, DIK_D , -1, 0, 0 },
		{ 1, DIK_S , -1, 0, 0 },
		{ 1, DIK_F , -1, 0, 0 },
		{ 1, DIK_RETURN , -1, 0, 0 },
		{ 1, DIK_X, -1, 0, 0 },
		{ 1, DIK_Z, -1, 0, 0 },
		{ 1, DIK_A, -1, 0, 0 },
		{ 1, DIK_SPACE, -1, 0, 0 },
		{ 1, DIK_Q, -1, 0, 0 },
		{ 1, DIK_R, -1, 0, 0 },

		{ 1, DIK_X, 1 | 0<<4 | 4<<8 | (160>>2)<<12 | (156>>2)<<18 | (40>>2)<<24 | 1<<30, 65535, 0 },
		{ 1, DIK_Z, 3 | 2<<4 | 6<<8 | (160>>2)<<12 | (156>>2)<<18 | (40>>2)<<24 | 1<<30, 65535, 0 },

		{ 1, DIK_D , 0x6800, 0, 0 },
		{ 1, DIK_S , 0x9400, 0, 0 },
		{ 1, DIK_F , 0, 0, 0 },
		{ 1, DIK_F , 0, 0, 0 },

		{ 1, DIK_T, 1, 0, 0 },
	}
};

static int change_pad_pushed[PGN_PADS] = { 0, 0 };
static int pad_type[PGN_PADS];
static int save_ini = 1;
static int save_ini_temp = 1;
static char *save_ini_section = "general";
static char *save_ini_key = "useinifile";

static char *ininame[PGN_PADS] = { "pgnneco.ini", "pgnneco2.ini" };
static char *section_str[NUM_KEY_CFGS] = { "DOWN", "UP", "RIGHT", "LEFT",
			"btn A", "btn B", "btn R", "btn START",
			"nejiri R", "nejiri L",
			"axis I", "axis II", "axis L",
			"btn Select", "btn L2", "btn R2",
			"af DOWN", "af UP", "af RIGHT", "af LEFT",
			"af btn A", "af btn B", "af btn R", "af btn START",
			"af axis I", "af axis II", "af axis L",
			"af btn Select", "af btn L2", "af btn R2",
			"bravo 1", "bravo 2",
			"bravo sub 1","bravo sub 2","bravo sub 3","bravo sub 4",
			"pad type"
 };

static char *key_str[] = { "device", "key", "analog", "max", "min" };
static char *axis_str[] = { "X", "Y", "Z", "RX", "RY", "RZ",
			"SL0", "SL1", "POV0", "POV1", "POV2", "POV3" };
static char *axis_sign[] = { "-", "+" };
static char *device_str[] = { "KB", "J1", "J2", "J3", "J4", "J5", "J6", "J7",
		"J8", "J9", "J10", "J11", "J12", "J13", "J14", "J15" };
static char *kb_str[256]={
	"", "Esc", "1", "2", "3", "4", "5", "6",
	 "7", "8", "9", "0", "-", "=", "BS", "Tab",
	"Q", "W", "E", "R", "T", "Y", "U", "I",
	 "O", "P", "[", "]", "Return", "L-Ctrl", "A", "S",
	"D", "F", "G", "H", "J", "K", "L", ";",
	 "'", "~", "L-Shift", PGN_BACKSLASH, "Z", "X", "C", "V",
	"B", "N", "M", "<", ">", "/", "R-Shift", "N*",
	 "L-Alt", "Space", "Caps", "F1", "F2", "F3", "F4", "F5",
	"F6", "F7", "F8", "F9", "F10", "Num", "Scroll", "N7",
	 "N8", "N9", "N-", "N4", "N5", "N6", "N+", "N1",
	"N2", "N3", "N0", "N.", "", "", "<(de)", "F11",
	 "F12", "", "", "", "", "", "", "",
	"", "", "", "", "F13", "F14", "F15", "",
	 "", "", "", "", "", "", "", "",
	"Kana", "", "", "/(br)", "", "", "", "",
	 "", "Henkan", "", "Muhenkan", "", PGN_YEN, "N.(br)", "",
	"", "", "", "", "", "", "", "",
	 "", "", "", "", "", "N=", "", "",
	"^", "@", ":", "_", "Kanji", "Stop", "AX", "(J3100)",
	 "", "M_Next", "", "", "NEnter", "R-Ctrl", "", "",
	"M_Mute", "Calc", "M_Play", "", "M_Stop", "", "", "",
	 "", "", "", "", "", "", "M_VolDn", "",
	"M_VolUp", "", "W_Home", "N,", "", "N/", "", "SysRq",
	 "R-Alt", "", "", "", "", "", "", "",
	"", "", "", "", "", "Pause", "", "Home",
	 "Up", "PgUp", "", "Left", "", "Right", "", "End",
	"Down", "PgDown", "Ins", "Del", "", "", "", "",
	 "", "", "", "L-Win", "R-Win", "Apps", "Power", "Sleep",
	"", "", "", "Wakeup", "", "W_Search", "W_Favorites", "W_Refresh",
	 "W_Stop", "W_Fwd", "W_Back", "M_MyComp", "Mail", "M_Sel", "", "",
	"", "", "", "", "", "", "", "",
	 "", "", "", "", "", "", "", ""
};

/*
 * DllMain()
 */
BOOL APIENTRY DllMain(HINSTANCE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	hinst = (HINSTANCE) hModule;
	return TRUE;
}

/* 
 * *current_dir_w32() adds absolute current path to filename.
 */
static char *current_dir_w32(char *result, char *filename)
{
	char dir[MAX_PATH];

        GetCurrentDirectory(MAX_PATH, dir);
	sprintf(result, "%s\\%s", dir, filename);
	return result;
}

/*
 * read_reg() reads registry.
 */
static int read_reg(char *section, char *key, int defvalue, int n)
{
	HKEY h;
	int data;
	int size;
	unsigned char buf[255];
	int result;
	int type;

	type = REG_DWORD;
	sprintf(buf,"p%d %s %s",n + 1, section, key);
	size = sizeof(data);
	result = RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Vision Thing\\PSEmu Pro\\PAD\\padGnneco", 0, KEY_READ, &h);
	if (result) {
		return defvalue;
	} 
	result = RegQueryValueEx(h, buf, NULL, (void*)&type, (void*)&data, (void*)&size);
	RegCloseKey(h);
	if (result) {
		return defvalue;
	} 

	return data;

}

/*
 * read_config_sub() reads a value from ini file or registry.
 */
static int read_config_sub(char * section, char *key , int default_value, int id, int use_ini)
{
	char fullpath[MAX_PATH];

	if (use_ini) {
		current_dir_w32(fullpath, ininame[id]);
		return (int) GetPrivateProfileInt(section, key, default_value, fullpath);
	} else {
		return read_reg(section, key, default_value, id);
	}
}

static int cfg2type(int config, int type)
{
	int i;
	if (type && (type < (1 << (MAX_PAD_BITS - 1)))) {
		type <<= 1;
	} else {
		type = 1;
	}
	for (i = 0; i < MAX_PAD_BITS; i++) {
		if ((type << i) < (1 << MAX_PAD_BITS)) {
			if (config & (type << i)) {
				return type << i;
			}	
		} else {
			if (config & (type >> (MAX_PAD_BITS - i))) {
				return type  >> (MAX_PAD_BITS - i);
			}
		}
	}
	return 1;
}

/*
 * read_config() reads and sets whole config.
 */
static void read_config(keyconfig k_cfg[NUM_KEY_CFGS], int id, int use_ini, const keyconfig k_default[])
{
	int i;

	for (i = 0; i < NUM_KEY_CFGS; i++) {
		k_cfg[i].device = read_config_sub(section_str[i], key_str[0], k_default[i].device, id, use_ini);
		k_cfg[i].key = read_config_sub(section_str[i], key_str[1], k_default[i].key, id,use_ini);
		k_cfg[i].analog = read_config_sub(section_str[i], key_str[2], k_default[i].analog, id, use_ini);
		k_cfg[i].max = read_config_sub(section_str[i], key_str[3], k_default[i].max, id, use_ini);
		k_cfg[i].min = read_config_sub(section_str[i], key_str[4], k_default[i].min, id, use_ini);
	}
	if (!k_cfg[NUM_KEY_CFGS-1].analog) {
		k_cfg[NUM_KEY_CFGS-1].analog = 1;
	}
	pad_type[id] = cfg2type(k_cfg[NUM_KEY_CFGS-1].analog, 0);
}

/*
 * ini_or_reg() decides ,returns 1 for ini, 0 for registry.
 */
static int ini_or_reg()
{
	int data;

	data = read_config_sub(save_ini_section, save_ini_key, -1, 0, 1);
	if (data == 1) return 1;
	data = read_config_sub(save_ini_section, save_ini_key, -1, 0, 0);
	if (data == 0) return 0;
	return 1;
}

/*
 * pgn_load_conf() reads all config.
 */
void pgn_load_conf()
{
	int i;
	int load_ini;

	load_ini = ini_or_reg();
	for (i = 0; i < PGN_PADS; i++) {
		read_config(key_config[i], i, load_ini, key_config_default[i != 0]);
	}
	save_ini = load_ini;
}

/*
 * write_reg() writes registry.
 */
static int write_reg(char * section,char * key,int value,int n)
{
	HKEY h;
	int size;
	unsigned char buf[255];
	int result;

	sprintf(buf,"p%d %s %s",n + 1, section, key);
	size = sizeof(value);
	result = RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Vision Thing\\PSEmu Pro\\PAD\\padGnneco", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &h, NULL);
	if (result) {
		RegCloseKey(h);
		return -1;
	} 
	result = RegSetValueEx(h, buf, 0, REG_DWORD,(BYTE *) &value, sizeof(value));
	if (result) {
	RegCloseKey(h);
		return -1;
	} 
	RegCloseKey(h);

	return 0;
}

/*
 * write_config_sub() writes a value to ini file or registry.
 */
static void write_config_sub(char * section, char *key, int value, int id, int use_ini)
{
	char fullpath[MAX_PATH];
	char c[20];

	if (use_ini) {
		current_dir_w32(fullpath, ininame[id]);
		sprintf (c, "%d", value);
		WritePrivateProfileString(section, key, c, fullpath);
	} else {
		write_reg(section, key, value, id);
	}
	return;
}

/*
 * write_config() writes whole config.
 */
static void write_config(keyconfig k_cfg[NUM_KEY_CFGS], int id, int use_ini)
{
	int i;

	for (i = 0; i < NUM_KEY_CFGS; i++) {
		write_config_sub(section_str[i], key_str[0], k_cfg[i].device, id, use_ini);
		write_config_sub(section_str[i], key_str[1], k_cfg[i].key, id, use_ini);
		write_config_sub(section_str[i], key_str[2], k_cfg[i].analog, id, use_ini);
		write_config_sub(section_str[i], key_str[3], k_cfg[i].max, id, use_ini);
		write_config_sub(section_str[i], key_str[4], k_cfg[i].min, id, use_ini);
	}
	write_config_sub(save_ini_section, save_ini_key, use_ini, id, use_ini);
	if (use_ini != ini_or_reg()) {
		write_config_sub(save_ini_section, save_ini_key, use_ini, id, !use_ini);
	}
}


/*
 * pgn_save_conf() writes all config.
 */
static void pgn_save_conf()
{
	int i;

	for (i = 0; i < PGN_PADS; i++) {
		write_config(key_config[i], i, save_ini);
	}
}

/*
 * release_di() releases all device and terminates directinput.
 */
static void release_di()
{ 
	int i;

	for (i = 0; i < PGN_MAX_DEV; i++) {
		if (di_dev[i]) { 
			di_dev[i]->lpVtbl->Unacquire(di_dev[i]);
			di_dev[i]->lpVtbl->Release(di_dev[i]);
			di_dev[i] = NULL;
		}
	}
	if (dinput) {
		dinput->lpVtbl->Release(dinput);
		dinput = NULL;
	}
}

/*
 * pgn_close() is called by PADclose() and terminates directinput.
 */
void pgn_close()
{
	release_di();
}

/*
 * init_joy() is called by EnumDevices in init_di(), sets guid and
 * increments num_dev when another joypad is found.
 */
BOOL CALLBACK init_joy(LPCDIDEVICEINSTANCE lpddi, int *num_dev)
{
	if (*num_dev < PGN_MAX_DEV) {
		guid[*num_dev] = lpddi->guidInstance;
		(*num_dev)++;
		return DIENUM_CONTINUE ;
	} else {
		return DIENUM_STOP ;
	}
}

/*
 * used_device() sums device to be used.
 */
const static int pad_keys[MAX_PAD_BITS][NUM_DLG_CFGS] = { { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE },
							 { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE },
							 { TRUE, TRUE, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE }};

static unsigned used_device(keyconfig *k_cfg, int id)
{
	int i,j,tmp;
	unsigned data = 0;

	if (k_cfg) {
		for (i = 0; i < NUM_KEY_CFGS - 1-2-4; i++) {
			if (i < NUM_DLG_CFGS || k_cfg[i].analog != -1 ) {
				for (j = 0; j < MAX_PAD_BITS; j++) {
					if (pad_keys[j][i] && k_cfg[NUM_KEY_CFGS-1].analog & (1<<j) ) {
						data |= k_cfg[i].device;
					}
				}
			}
		}
		tmp = cfg2type(k_cfg[NUM_KEY_CFGS - 1].analog, 0);
		if (tmp != cfg2type(k_cfg[NUM_KEY_CFGS - 1].analog, tmp)) {
			data |= k_cfg[NUM_KEY_CFGS - 1].device;
		}
		if (k_cfg[NUM_KEY_CFGS - 1].analog & 4) { //bravo
			for (i = NUM_DLG_CFGS + NUM_DLG_AF; i < NUM_DLG_CFGS + NUM_DLG_AF + 2; i++) {
				data |= k_cfg[i].device;
			}
			for (i = NUM_DLG_CFGS + NUM_DLG_AF + 2; i < NUM_DLG_CFGS + NUM_DLG_AF + 2 + 4; i++) {
				if (k_cfg[i].analog) {
					data |= k_cfg[i].device;
				}
			}
		}
	} else {
		for (i = 0; i < PGN_MAX_DEV; i++) {
			data |= 1 << i;
		}
	}
	return data;
}

/*
 * di_err() reports error.
 */
static void di_err(HRESULT hr)
{
	char s[100];

	switch (hr) {
	case DIERR_BETADIRECTINPUTVERSION:
		MessageBox(NULL, "Beta", LIB_NAME, MB_ICONERROR);
		break;
	case DIERR_INVALIDPARAM:
		MessageBox(NULL, "Invalid param", LIB_NAME, MB_ICONERROR);
		break;
	case DIERR_OLDDIRECTINPUTVERSION:
		MessageBox(NULL, "Old DirectInput", LIB_NAME, MB_ICONERROR);
		break;
	case DIERR_OUTOFMEMORY:
		MessageBox(NULL, "Out of memory", LIB_NAME, MB_ICONERROR);
		break;
	default:
		sprintf (s, "Error hr = %d", (int) hr);
		MessageBox (NULL, s, LIB_NAME, MB_ICONERROR);
		break;
	}
}

/*
 * di_acquire() gets access to device.
 */
static HRESULT di_acquire(PGN_LPDIRECTINPUTDEVICE di_dev)
{
	HRESULT hr;

	hr = di_dev->lpVtbl->Acquire(di_dev);
	return hr;
}

#define PGN_ERROR_CHECK if (hr < 0) { di_err(hr); return FALSE; }

/*
 * init_di() initializes directinput and devices to be used.
 */
static BOOL init_di(HWND hwndMain, keyconfig *k_cfg, int id)
{
	HRESULT hr; 
	int i = 0;
	unsigned tmp;
	static int num_dev;

	if (!dinput) {
		for (i = 0; i < PGN_MAX_DEV; i++) {
			di_dev[i] = NULL;
		}
#if DIRECTINPUT_VERSION >= 0x0800
		hr = DirectInput8Create(hinst, DIRECTINPUT_VERSION, &IID_IDirectInput8, (void *) &dinput, NULL); 
#else
		hr = DirectInputCreate(hinst, DIRECTINPUT_VERSION, (void *) &dinput, NULL); 
#endif
		PGN_ERROR_CHECK
		num_dev = 1;
		guid[0] = GUID_SysKeyboard;
#if DIRECTINPUT_VERSION >= 0x0800
		hr = dinput->lpVtbl->EnumDevices(dinput, DI8DEVCLASS_GAMECTRL, (LPDIENUMDEVICESCALLBACK) init_joy, &num_dev, DIEDFL_ATTACHEDONLY);
#else
		hr = dinput->lpVtbl->EnumDevices(dinput, DIDEVTYPE_JOYSTICK, (LPDIENUMDEVICESCALLBACK) init_joy, &num_dev, DIEDFL_ATTACHEDONLY);
#endif
		PGN_ERROR_CHECK
	}
	tmp = used_device(k_cfg, id);
	for (i = 0; i < num_dev; i++) {
		if (di_dev[i]) break;
		if (tmp & (1 << i)) {
			hr = dinput->lpVtbl->CreateDevice(dinput, &guid[i], (void *) &di_dev[i], NULL); 
			PGN_ERROR_CHECK
			if (i) {
				hr = di_dev[i]->lpVtbl->SetDataFormat(di_dev[i],&c_dfDIJoystick);
			} else {
				hr = di_dev[0]->lpVtbl->SetDataFormat(di_dev[0],&c_dfDIKeyboard);
			}
			PGN_ERROR_CHECK
			if (i) {
				hr = di_dev[i]->lpVtbl->SetCooperativeLevel(di_dev[i], hwndMain, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE);
			} else {
#if DI_VER < 7
				hr = di_dev[i]->lpVtbl->SetCooperativeLevel(di_dev[i], hwndMain, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
#else
				hr = di_dev[0]->lpVtbl->SetCooperativeLevel(di_dev[0], hwndMain, DISCL_FOREGROUND | (k_cfg ? DISCL_NONEXCLUSIVE : DISCL_EXCLUSIVE));
#endif
			}
			PGN_ERROR_CHECK
			hr = di_acquire(di_dev[i]);
			PGN_ERROR_CHECK
		}
	}
	return TRUE;
}

/*
 * pgn_open() is calld by PADopen() and calls initialize function.
 */
void pgn_open(void *Disp, int i, int ps2e)
{
	BOOL test;

	if (!ps2e) {
		test = init_di((HWND) Disp, key_config[i], i); //pse
	} else {
		test = init_di((HWND) *(long*) Disp, key_config[i], i); //ps2e
	}
	if (test != TRUE) {
		release_di();
	}
}

/*
 * key_data() caluculates a key data from inputstate and keyconfig.
 */
static int key_data(int analog, inputstates *i_st, keyconfig *k_cfg)
{
	void *p;
	int tmp;
	unsigned device;

	for (device = 0; device < PGN_MAX_DEV; device++) {
		if (k_cfg->device & (1 << device)) break;
	}
	if (device) {
		p = (void*) &i_st->j_state[device-1];
	} else {
		p = (void*) i_st->k_state;
	}
	if ((k_cfg->key < DIJOFS_BUTTON0) && device) { // axis
		tmp = *(int*) (p + (k_cfg->key));
		if ((k_cfg->min < k_cfg->max && k_cfg->min < tmp) || (k_cfg->min > k_cfg->max && tmp < k_cfg->min)) {
			if (analog) {
				return min(analog, (tmp - k_cfg->min) * analog / (k_cfg->max - k_cfg->min));
			} else {
				return 1;
			}
		}
	} else { // btn, keyboard
		if (analog) {
			return (*((char*)p + (k_cfg->key)) & 0x80) ? analog : 0;
		} else {
			return (*((char*)p + (k_cfg->key)) & 0x80) != 0;
		}
	}
	return 0;
}

/*
 * update_devices() polls device by needs
 */
static int update_devices(inputstates *i_st)
{
	int i;
	HRESULT hr;

	for (i = 0; i < PGN_MAX_DEV; i++) {
		if (di_dev[i]) {
			hr = di_dev[i]->lpVtbl->Poll(di_dev[i]);
			if (hr < 0) {
				if (hr == DIERR_NOTACQUIRED || hr == DIERR_INPUTLOST) {
					di_acquire(di_dev[i]);
					di_dev[i]->lpVtbl->Poll(di_dev[i]);
				} else {
					return -1;
				}
			}
			if (i) {
				hr = di_dev[i]->lpVtbl->GetDeviceState(di_dev[i], sizeof(DIJOYSTATE), &i_st->j_state[i-1]);
			} else {
				hr = di_dev[i]->lpVtbl->GetDeviceState(di_dev[i], 256, &i_st->k_state);
				if (hr < 0) {
					memset(i_st->k_state,0,256);
				}
			}
		}
	}
	return 0;
}

/*
 * autofire2() returns 1 (push) or -1 (free) if autofire is enabled.
 */
static int autofire2(inputstates *i_st, keyconfig k_cfg[NUM_KEY_CFGS], int id, int i)
{
	if (k_cfg[i + NUM_DLG_CFGS].analog != -1) {
		if (key_data(0, i_st, &k_cfg[i + NUM_DLG_CFGS])) {
			if (af_timer[id][i] >= k_cfg[i + NUM_DLG_CFGS].analog * 2) {
				af_timer[id][i] = 0;
			}
			if (af_timer[id][i] < k_cfg[i + NUM_DLG_CFGS].analog || af_timer[id][i] == 0) {
				af_timer[id][i]++;
				return 1;
			} else {
				af_timer[id][i]++;
				return -1;
			}
		} else {
			af_timer[id][i] = 0;
		}
	}
	return 0;
}

typedef struct {char * string; unsigned btn;} btn_def;
btn_def bdef[]={
	{ STR_MARU, 1<<13 },
	{ STR_BATSU, 1<<14 },
	{ STR_SANKAKU, 1<<12 },
	{ STR_SHIKAKU, 1<<15 },
	{ "R1", 1<<11 },
	{ "R2", 1<<9 },
	{ "L1", 1<<10 },
	{ "L2",  1<<8 }
};

/*
 * read_pad() calculates pad data.
 */
unsigned char read_pad(unsigned char buf[6], inputstates *i_st, keyconfig k_cfg[NUM_KEY_CFGS], int id)
{
	const unsigned btn[14] = { 1<<6, 1<<4, 1<<5, 1<<7, 1<<13, 1<<12, 1<<11, 1<<3, 1<<14, 1<<15, 1<<10, 1<<0, 1<<8, 1<<9 };
	static unsigned char prev_bravo[2][2];
	unsigned char bravo[2];
	static int b_val[2][2];
	static int update_flag = 0xff;
	unsigned data = 0xffff;
	int tmp;
	int i;
	int flag;

	flag = 1 << id;
	if (update_flag & flag) {
		if (update_devices(i_st)) {
			switch (pad_type[id]) {
			case 1:
				buf[0]= 0xff;
				buf[1]= 0xff;
				buf[2]= 0x80;
				buf[3]= 0;
				buf[4]= 0;
				buf[5]= 0;
				return 0x23;
			case 2:
			case 4:
			default:
				buf[0]= 0xff;
				buf[1]= 0xff;
				buf[2]= 0xff;
				buf[3]= 0xff;
				buf[4]= 0xff;
				buf[5]= 0xff;
				return 0x41;
			}
		}
		update_flag = flag;
	} else {
		update_flag |= flag;
	}
	if (change_pad_pushed[id] && !key_data(0, i_st, &k_cfg[NUM_KEY_CFGS-1])) {
		pad_type[id] = cfg2type(k_cfg[NUM_KEY_CFGS-1].analog, pad_type[id]);
		change_pad_pushed[id] = 0;
	}
	tmp = cfg2type(k_cfg[NUM_KEY_CFGS-1].analog, 0);
	if (tmp != cfg2type(k_cfg[NUM_KEY_CFGS-1].analog, tmp) && key_data(0, i_st, &k_cfg[NUM_KEY_CFGS-1])) {
		change_pad_pushed[id] = 1;
	}
	switch (pad_type[id]) {
	case 1:
		for (i = 0; i < 8; i++) {
			if (key_data(0, i_st, &k_cfg[i])) {
				data &= ~btn[i];
			}
			switch (autofire2(i_st, k_cfg, id, i)) {
			case 1:
				data &= ~btn[i];
				break;
			case -1:
				data |= btn[i];
				break;
			}
		}
		buf[2] = 0x80 + key_data(min(127, k_cfg[8].analog), i_st, &k_cfg[8]) - key_data(min(128, k_cfg[9].analog), i_st, &k_cfg[9]);
		for (i = 10; i < NUM_NEG; i++) {
			buf[i-7] = key_data(min(255, k_cfg[i].analog), i_st, &k_cfg[i]);
			switch (autofire2(i_st, k_cfg, id, i-2)) {
			case 1:
				buf[i-7] = 255;
				break;
			case -1:
				buf[i-7] = 0;
				break;
			}
		}
		buf[0] = data & 0xff;
		buf[1] = data >> 8 & 0xff;
		return 0x23;
	case 4:
		for (i = 0; i < 14; i++) {
			if(pad_keys[2][i<8?i:i+2]) {
				if (key_data(0, i_st, &k_cfg[i<8?i:i+2])) {
					data &= ~btn[i];
				}
			}
			switch (autofire2(i_st, k_cfg, id, i)) {
			case 1:
				data &= ~btn[i];
				break;
			case -1:
				data |= btn[i];
				break;
			}
		}
		for (i = 0; i < 2; i++) {
			if (k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (1<<30)){
				bravo[i] = key_data(255, i_st, &k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i]);
				if (bravo[i] > (k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<12)) >> (12-2)) {
					bravo[i] = (k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<12)) >> (12-2);
					if (!b_val[id][i]) {
						tmp = bravo[i] - prev_bravo[id][i];
						if (tmp>((k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<18)) >> (18-2))) {
							b_val[id][i] = 3;
						}else if (tmp>((k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<24)) >> (24-2))) {
							b_val[id][i] = 2;
						} else {
							b_val[id][i] = 1;
						}
					}
				} else {
					b_val[id][i] = 0;
				}
				prev_bravo[id][i] = bravo[i];
			} else {
				tmp = key_data(255, i_st, &k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i]);
				if (!b_val[id][i]) {
					if(tmp>((k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<12)) >> (12-2))){
						b_val[id][i] = 3;
					} else if (tmp>((k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<18)) >> (18-2))) {
						b_val[id][i] = 2;
					} else if (tmp>((k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<24)) >> (24-2))) {
						b_val[id][i] = 1;
					}
				} else if (tmp<=((k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog & (0x3f<<24)) >> (24-2))){
					b_val[id][i] = 0;
				}
			}
		}
		for (i = 0; i < 2; i++) {
			if (b_val[id][i]) {
				data &= ~bdef[(0x0f << ((b_val[id][i] - 1) * 4) & (k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + i].analog)) >> ((b_val[id][i] - 1) * 4)].btn;
			}
		}
		for (i = 0; i < 4; i++) {
			if (k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + 2 + i].analog) {
				if (key_data(0, i_st, &k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + 2 +i])) {
					data &= ~(unsigned)k_cfg[NUM_DLG_CFGS + NUM_DLG_AF + 2 + i].analog;
				}
			}
		}
		buf[0] = data & 0xff;
		buf[1] = data >> 8 & 0xff;
		return 0x41;
	case 2:
	default:
		for (i = 0; i < 8; i++) {
			if (key_data(0, i_st, &k_cfg[i])) {
				data &= ~btn[i];
			}
			switch (autofire2(i_st, k_cfg, id, i)) {
			case 1:
				data &= ~btn[i];
				break;
			case -1:
				data |= btn[i];
				break;
			}
		}
		for (i = 0; i < 3; i++) {
			if (key_data(min(255, k_cfg[i+10].analog), i_st, &k_cfg[i+10]) > 0x20) {
				data &= ~btn[i+8];
			}
			switch (autofire2(i_st, k_cfg, id, i+8)) {
			case 1:
				data &= ~btn[i+8];
				break;
			case -1:
				data |= btn[i+8];
				break;
			}
		}
		for (i = 0; i < 3; i++) {
			if (key_data(0, i_st, &k_cfg[i+13])) {
				data &= ~btn[i+11];
			}
			switch (autofire2(i_st, k_cfg, id, i+11)) {
			case 1:
				data &= ~btn[i+11];
				break;
			case -1:
				data |= btn[i+11];
				break;
			}
		}
		buf[0] = data & 0xff;
		buf[1] = data >> 8 & 0xff;
		return 0x41;
	}
}

/*
 * pgn_read_port() is called by PADreadPort1() and PADreadPort2() and  * reports data for a pad.
 */
unsigned char pgn_read_port(unsigned char buf[6], int id)
{
	return read_pad(buf, &input_state, key_config[id], id);
}

/*
 * set_obj_text_w32() sets text of gui object.
 */
static void set_obj_text_w32(HWND hw, WPARAM w, char *c)
{
	HWND obj;
	obj = GetDlgItem(hw, w);
	SetWindowText(obj, c);
	return;
}

/*
 * set_obj_text_w32() enable/disables gui object.
 */
static void enable_obj_w32(HWND hw, WPARAM w, BOOL b)
{
	HWND obj;
	obj = GetDlgItem(hw, w);
	EnableWindow(obj, b);
	return;
}

/*
 * set_btn_text() calculate and shows string on gui bottun from config.
 */
static void set_btn_text(HWND hw, int id, keyconfig *k_cfg)
{
	char btn_text[30];
	int i;

	for (i = 0; i < PGN_MAX_DEV; i++) {
		if ((1 << i) & k_cfg->device) {
			break;
		}
	}
	if (i) {
		if (k_cfg->key < DIJOFS_BUTTON0) { // joy axis
			sprintf(btn_text, "%s %s %s", device_str[i], axis_str[(k_cfg->key - (int) DIJOFS_X) / sizeof(long)], axis_sign[k_cfg->max > k_cfg->min]);
		} else { // joy btn
			sprintf(btn_text, "%s btn %d", device_str[i],(k_cfg->key - (int) DIJOFS_BUTTON0));
		}
	} else { // keyboard
		sprintf(btn_text, "%s %s", device_str[i], kb_str[k_cfg->key]);
	}
	set_obj_text_w32(hw, id, btn_text);
}


/*
 * compare_states() compares two input states and reports device
 * constant to be configured.
 */
static int compare_states(inputstates *i_st0, inputstates *i_st1, keyconfig cfg_candidate[PGN_MAX_CANDIDATES])
{
	int i;
	int j = 0;
	int device = 0;

	for (device = 0; device < PGN_MAX_DEV; device++) {
		if (device < PGN_MAX_DEV && di_dev[device]) {
			if (!device) {
				for (i = 0; i < 256; i++) {
					if ((i_st1->k_state[i] & 0x80) != (i_st0->k_state[i] & 0x80)) {
						if (j < PGN_MAX_CANDIDATES && cfg_candidate) {
							cfg_candidate[j].device = 1 << device;
							cfg_candidate[j].key = i;
						}
					j++;
					}
				}
			} else {
				for (i = 0; i < 32; i++) {
					if (i_st1->j_state[device-1].rgbButtons[i] != i_st0->j_state[device-1].rgbButtons[i]) {
						if (j<PGN_MAX_CANDIDATES && cfg_candidate) {
							cfg_candidate[j].device = 1 << device;
							cfg_candidate[j].key = i + DIJOFS_BUTTON0;
						}
					j++;
					}
				}
				for (i = 0; i < 8; i++) {
					if (abs( *(&i_st1->j_state[device-1].lX+i) - *(&i_st0->j_state[device-1].lX+i)) > threshold) {
						if (j<PGN_MAX_CANDIDATES && cfg_candidate) {
							cfg_candidate[j].device = 1 << device;
							cfg_candidate[j].key = i * sizeof(long) + DIJOFS_X;
							cfg_candidate[j].max = *(&i_st1->j_state[device-1].lX+i);
							cfg_candidate[j].min = *(&i_st0->j_state[device-1].lX+i);
						}
					j++;
					}
				}
			}
		}
	}
	return j;
}

/*
 * ChooseBtnDlgProc() executes "choose a button" dialog, sets a config
 * from several candidates.
 */
LRESULT CALLBACK ChooseBtnDlgProc(HWND hd, UINT msg, WPARAM w, LPARAM l)
{
	switch (msg) {
	case WM_INITDIALOG:
	{
		int i;
		for (i = 0; i < PGN_MAX_CANDIDATES; i++) {
			if (key_config_candidate[i].device) {
				set_btn_text(hd, ID_B1 + i, &key_config_candidate[i]);
				enable_obj_w32(hd,ID_B1 + i, TRUE);
			}
		}
	}
		break;
	case WM_COMMAND:
		switch (LOWORD(w)) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hd, w);
			return TRUE;
		default:
			if (ID_B1 <= LOWORD(w) && LOWORD(w) < ID_B1 + PGN_MAX_CANDIDATES) {
				EndDialog(hd, w);
				return TRUE;
			}
		}
	}
	return FALSE;
}

/*
 * set_config() sets one key config.
 */
static int set_config(keyconfig *k_in, keyconfig *k_cfg, int digital)
{
	k_cfg->device = k_in->device;
	k_cfg->key = k_in->key;
	if ((k_cfg->device >> 1) && (k_cfg->key < DIJOFS_BUTTON0)) {
		k_cfg->max = k_in->max;
		if (digital) {
			k_cfg->min = (k_cfg->max + k_in->min) / 2;
		} else {
			k_cfg->min = k_in->min;
		}
	} else {
		k_cfg->max = 0;
		k_cfg->min = 0;
	}
	return 0;
}

/*
 * AboutDlgProc() shows about popup.
 */
BOOL CALLBACK AboutDlgProc(HWND hw, UINT msg, WPARAM w, LPARAM l)
{
	switch (msg) {
		case WM_COMMAND: {
			switch (LOWORD(w)) {
			case IDOK:
			case IDCANCEL:
				EndDialog(hw, TRUE);
				return TRUE;
			}
		}
	}
	return FALSE;
}

/*
 * pgn_about() is called by PADabout() and calls dialog procedure.
 */
void pgn_about()
{
	DialogBox(hinst, MAKEINTRESOURCE(IDPGN_ABOUT_DLG), GetActiveWindow(), (DLGPROC)AboutDlgProc);
}

/*
 * get_threshold() converts combobox to int value.
 */
#define PGN_MAX_FIGURES 12
int get_threshold(HWND hw, int id)
{
	HWND h;
	char c[PGN_MAX_FIGURES + 1];

	h = GetDlgItem(hw, id);
	GetWindowText(h, c, PGN_MAX_FIGURES);
	return atoi(c);
}

/*
 * init_update_release() automatically opens, reads, and closes directinput.
 */
int init_update_release(HWND hw, DWORD time, inputstates *i_st)
{
	if (init_di(hw, NULL, 0) != TRUE) {
		release_di();
		return -1;
	}
	Sleep(time);
	update_devices(i_st);
	release_di();
	Sleep(time);
	return 0;
}

/*
 * choose_config() chooses and enables suitable config.
 */
static void choose_config(int btn_n, int btn_id, keyconfig cfg_candidate[PGN_MAX_CANDIDATES], keyconfig cfg_dialog[2][NUM_KEY_CFGS], HWND hw,int digital)
{
	int i;

	if (cfg_candidate[1].device) {
		i = LOWORD(DialogBox(hinst, MAKEINTRESOURCE(IDPGN_CHOOSEBTN) ,hw , (DLGPROC) ChooseBtnDlgProc));
		switch (i) {
		case IDOK:
		case IDCANCEL:
			break;
		default:
			if (ID_B1 <= i && i < ID_B1 + PGN_MAX_CANDIDATES) {
				set_config(&cfg_candidate[i - ID_B1], &cfg_dialog[btn_n][btn_id], digital);
			}
		}
	} else if (cfg_candidate[0].device) {
		set_config(&cfg_candidate[0], &cfg_dialog[btn_n][btn_id], digital);
	}
}

/*
 * scan_input() scans and sets a input for a button.
 */
static void scan_input(HWND hw, WPARAM w)
{
	int i;

	if (!pushed) {
		if (!init_update_release(hw, 10, &input_state)) {
			pushed = 1;
			Sleep(100);
		} else return;
	}
	if (init_di(hw, NULL, 0) != TRUE) {
		release_di();
		return;
	}
	threshold = get_threshold(hw, ID_CFG_THRESHOLD);
	enable_obj_w32(hw, w, FALSE);
	update_devices(&input_states_conf[0]);
	for (i = 1; i < (WAIT1 / DELAY1); i++) {
		update_devices(&input_states_conf[0]);
		if (compare_states(&input_states_conf[0], &input_state, NULL)) break;
		Sleep(DELAY1);
	}
	update_devices(&input_states_conf[0]);
	Sleep(DELAY1 + DELAY2);
	for (i = 1; i < (WAIT2 / DELAY2); i++) {
		update_devices(&input_states_conf[i % 2]);
		if (!compare_states(&input_states_conf[0], &input_states_conf[1], NULL)) break;
		Sleep(DELAY2);
	}
	memset(key_config_candidate, 0, sizeof(key_config_candidate));
	compare_states(&input_state, &input_states_conf[1], key_config_candidate);
	release_di();
}

LRESULT CALLBACK BravoDlgProc(HWND hw, UINT msg, WPARAM w, LPARAM l)
{
	HWND h;
	int i,j,k,id;

	const unsigned btn[14] = { 1<<6, 1<<4, 1<<5, 1<<7, 1<<13, 1<<12, 1<<11, 1<<3, 1<<14, 1<<15, 1<<10, 1<<0, 1<<8, 1<<9 };

	switch (msg) {
	case WM_INITDIALOG:
			for (id=0;id<2;id++){
				for (k=0;k<2;k++) {
					set_btn_text(hw, k+ID_BRAVO_SWITCH +id*2, &key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+k]);
					for (i=0;i<3;i++){
						h = GetDlgItem(hw, ID_BRAVO_LOW+i*10+k+id*2);
						for(j=0;j<8;j++){
							SendMessage(h, CB_ADDSTRING, 0, (LPARAM) bdef[j].string);
						}
						SendMessage(h, CB_SETCURSEL, ((0x0f<<(i*4))&(key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+k].analog))>>(i*4),0);
					}
					for (i=0;i<3;i++){
						h = GetDlgItem(hw, ID_BRAVO_VALUE_LOW+i*10+k+id*2);
						for(j=0;j<255;j+=(1<<2)){
							char c[10];
							sprintf(c,"%d",j);
							SendMessage(h, CB_ADDSTRING, 0, (LPARAM) c);
						}
						SendMessage(h, CB_SETCURSEL, ((0x3f<<(24-i*6))&(key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+k].analog))>>(24-i*6),0);
					}
					h = GetDlgItem(hw, ID_BRAVO_CHECK+k+id*2);
					SendMessage(h, BM_SETCHECK, (key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+k].analog & (1<<30)) ? BST_CHECKED : BST_UNCHECKED, 0);
				}
				for (k=0;k<4;k++) {
					set_btn_text(hw, k+ID_BRAVO_EXTRA_BTN +id*4, &key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+2+k]);
					if(key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+2+k].analog){
						enable_obj_w32(hw,k+ID_BRAVO_EXTRA_BTN +id*4,TRUE);
						for (i=0;i<14;i++){
							if (key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+2+k].analog & btn[i]){
								h = GetDlgItem(hw, ID_BRAVO_EXTRA_BTN_CHECK+k*20+id*80+i);
								SendMessage(h, BM_SETCHECK, BST_CHECKED, 0);
							}
						}
					}
				}
			}
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(w)) {
		case IDOK:
		case IDCANCEL:
			for (id=0;id<2;id++){
				for (k=0;k<2;k++) {
					for (i=0;i<3;i++){
						h = GetDlgItem(hw, ID_BRAVO_LOW+i*10+k+id*2);
						key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF + k].analog &= ~(0x0f << (4 * i));
						key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF + k].analog |= (SendMessage(h, CB_GETCURSEL, 0 ,0) & 0x0f) << (4 * i);
					}
					h = GetDlgItem(hw, ID_BRAVO_CHECK + k + id * 2);
					key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF + k].analog &= ~(1<<30);
					key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF + k].analog |= (SendMessage(h, BM_GETCHECK, 0, 0) == BST_CHECKED) ? (1<<30) : 0;
					for (i=0;i<3;i++){
						h = GetDlgItem(hw, ID_BRAVO_VALUE_LOW+i*10+k+id*2);
						key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF + k].analog &= ~(0x3f<<(24-6*i));
						key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF + k].analog |= (SendMessage(h, CB_GETCURSEL, 0 ,0) & 0x3f) << (24-6*i);
					}
				}
			}
			EndDialog(hw, w);
			return TRUE;
		default:
			if (ID_BRAVO_SWITCH <= w && w < ID_BRAVO_SWITCH + 2) {
				scan_input(hw,w);
				choose_config((w - ID_BRAVO_SWITCH) / 2, NUM_DLG_CFGS + NUM_DLG_AF+(w - ID_BRAVO_SWITCH) % 2, key_config_candidate, key_config_tmp, hw, 0 );
				set_btn_text(hw, w , &key_config_tmp[(w - ID_BRAVO_SWITCH) / 2][NUM_DLG_CFGS + NUM_DLG_AF+(w - ID_BRAVO_SWITCH) % NUM_DLG_CFGS]);
				enable_obj_w32(hw, w, TRUE);
			}
			if (ID_BRAVO_EXTRA_BTN <= w && w < ID_BRAVO_EXTRA_BTN + 4) {
				scan_input(hw,w);
				choose_config((w - ID_BRAVO_EXTRA_BTN) / 4, NUM_DLG_CFGS + NUM_DLG_AF + 2 +(w - ID_BRAVO_EXTRA_BTN) % 4, key_config_candidate, key_config_tmp, hw, 0 );
				set_btn_text(hw, w , &key_config_tmp[(w - ID_BRAVO_EXTRA_BTN) / 4][NUM_DLG_CFGS + NUM_DLG_AF+2+(w - ID_BRAVO_EXTRA_BTN) % NUM_DLG_CFGS]);
				enable_obj_w32(hw, w, TRUE);
			}
			if (ID_BRAVO_EXTRA_BTN_CHECK <= w && w < ID_BRAVO_EXTRA_BTN_CHECK + 140+14) {
				h = GetDlgItem(hw, w);
				if (SendMessage(h, BM_GETCHECK, 0, 0) == BST_CHECKED){
					key_config_tmp[(w - ID_BRAVO_EXTRA_BTN_CHECK) / 80][NUM_DLG_CFGS + NUM_DLG_AF+2+((w - ID_BRAVO_EXTRA_BTN_CHECK) % 80)/20].analog |= btn[((w - ID_BRAVO_EXTRA_BTN_CHECK) % 80)%20];
				} else {
					key_config_tmp[(w - ID_BRAVO_EXTRA_BTN_CHECK) / 80][NUM_DLG_CFGS + NUM_DLG_AF+2+((w - ID_BRAVO_EXTRA_BTN_CHECK) % 80)/20].analog &= ~btn[((w - ID_BRAVO_EXTRA_BTN_CHECK) % 80)%20];
				}
				for (id=0;id<2;id++){
					for (k=0;k<4;k++) {
						enable_obj_w32(hw,k+ID_BRAVO_EXTRA_BTN +id*4,(key_config_tmp[id][NUM_DLG_CFGS + NUM_DLG_AF+2+k].analog)?TRUE:FALSE);
					}
				}
			}
		}
	}
	return FALSE;
}

#define AF_SPEED_DEFAULT 2

/*
 * AutofireDlgProc() is autofire dialog procedure.
 */
LRESULT CALLBACK AutofireDlgProc(HWND hw, UINT msg, WPARAM w, LPARAM l)
{
	int i,j,k;
	HWND h;
	char c[5+1];
	const int af_list[] = { 0, 1, 2, 3, 5, 6, -1 };

	switch (msg) {
	case WM_INITDIALOG:
		for (j = 0; j < 2; j++) {
			for (i = 0; i < NUM_DLG_AF; i++) {
				set_btn_text(hw, i+ID_AF_BTN+j*(NUM_DLG_AF), &key_config_tmp[j][i+NUM_DLG_CFGS]);
				h = GetDlgItem(hw, i + ID_AF_SPEED + j * NUM_DLG_AF);
				for (k = 0; af_list[k] != -1; k++) {
					sprintf(c, "%d", af_list[k]);
					SendMessage(h, CB_ADDSTRING, 0, (LPARAM) c);
				}
				if (key_config_tmp[j][i+NUM_DLG_CFGS].analog != -1) {
					SendMessage(GetDlgItem(hw, i + ID_AF_ENABLE + j * NUM_DLG_AF), BM_SETCHECK, BST_CHECKED, 0);
					h = GetDlgItem(hw, i + ID_AF_SPEED + j * NUM_DLG_AF);
					sprintf(c, "%d", key_config_tmp[j][i + NUM_DLG_CFGS].analog);
					SetWindowText(h, c);
					enable_obj_w32(hw,i + ID_AF_BTN + j * NUM_DLG_AF, TRUE);
					enable_obj_w32(hw,i + ID_AF_SPEED + j * NUM_DLG_AF, TRUE);
				}
			}
		}
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(w)) {
		case IDOK:
			for (j = 0; j < 2; j++) {
				for (i = 0; i < NUM_DLG_AF; i++) {
					h = GetDlgItem(hw, i + ID_AF_SPEED + j * NUM_DLG_AF);
					GetWindowText(h, c, 5);
					key_config_tmp[j][i + NUM_DLG_CFGS].analog = atoi(c);
					if (!(SendMessage(GetDlgItem(hw, i + ID_AF_ENABLE + j * NUM_DLG_AF), BM_GETCHECK, 0, 0) == BST_CHECKED)) {
						key_config_tmp[j][i + NUM_DLG_CFGS].analog = -1;
					}
				}
			}
		case IDCANCEL:
			EndDialog(hw, w);
			return TRUE;
		default:
			if (ID_AF_ENABLE <= LOWORD(w) && LOWORD(w) < ID_AF_ENABLE + NUM_DLG_AF * 2) {
				int temp;
				temp = (SendMessage(GetDlgItem(hw, w), BM_GETCHECK, 0, 0) == BST_CHECKED);
				if (temp) {
					enable_obj_w32(hw,(w - ID_AF_ENABLE) + ID_AF_BTN, TRUE);
					enable_obj_w32(hw,(w - ID_AF_ENABLE) + ID_AF_SPEED, TRUE);
					h = GetDlgItem(hw, (w - ID_AF_ENABLE) + ID_AF_SPEED);
					GetWindowText(h, c, 5);
					if (c[0] == '\0') {
						sprintf(c, "%d", AF_SPEED_DEFAULT);
						SetWindowText(h, c);
					}
				} else {
					enable_obj_w32(hw, w - ID_AF_ENABLE + ID_AF_BTN, FALSE);
					enable_obj_w32(hw, w - ID_AF_ENABLE + ID_AF_SPEED, FALSE);
				}
				break;
			}
			if (ID_AF_BTN <= LOWORD(w) && LOWORD(w) < NUM_DLG_AF * 2 + ID_AF_BTN) {
				scan_input(hw,w);
				choose_config((w - ID_AF_BTN) / NUM_DLG_AF,(w - ID_AF_BTN) % NUM_DLG_AF + NUM_DLG_CFGS, key_config_candidate, key_config_tmp, hw, 1);
				set_btn_text(hw, w, &key_config_tmp[(w - ID_AF_BTN) / NUM_DLG_AF][(w - ID_AF_BTN) % NUM_DLG_AF + NUM_DLG_CFGS]);
				enable_obj_w32(hw, w, TRUE);
				break;
			}
		}
	}
	return FALSE;
}

/*
 * refresh_btns() updates texts on dialog.
 */
static void refresh_btns(HWND hw, int j)
{
	int i, tmp, k;

	for (i = 0; i < NUM_DLG_CFGS; i++) {
		tmp = 0;
		for (k = 0; k < MAX_PAD_BITS; k++) {
			if (1<<k & key_config_tmp[j][NUM_KEY_CFGS-1].analog) {
				tmp |= pad_keys[k][i];
			}
		}
		enable_obj_w32(hw, ID_BTN_DOWN + i + NUM_DLG_CFGS * j, tmp);
	}
	tmp = cfg2type(key_config_tmp[j][NUM_KEY_CFGS-1].analog, 0);
	enable_obj_w32(hw, j + ID_CHANGE_PAD, (tmp != cfg2type(key_config_tmp[j][NUM_KEY_CFGS-1].analog, tmp)) ? TRUE : FALSE);
}

/*
 * ConfigDlgProc() executes configure dialog, reads devices and saves
 * config.
 */
BOOL CALLBACK ConfigDlgProc(HWND hw, UINT msg, WPARAM w, LPARAM l)
{
	int i, j;
	int tmp;

	switch (msg) {
	case WM_INITDIALOG:
		for (j = 0; j < 2; j++) {
			for (i = 0; i < NUM_KEY_CFGS; i++) {
				key_config_tmp[j][i] = key_config[j][i];
			}
		}
		save_ini_temp = save_ini;
		ShowWindow(hw, SW_SHOW);
		for (j = 0; j < 2; j++) {
			for (i = 0; i < NUM_DLG_CFGS; i++) {
				set_btn_text(hw, j * NUM_DLG_CFGS + i + ID_BTN_DOWN, &key_config_tmp[j][i]);
			}
			set_btn_text(hw, j + ID_CHANGE_PAD, &key_config_tmp[j][NUM_KEY_CFGS-1]);
			tmp = cfg2type(key_config_tmp[j][NUM_KEY_CFGS-1].analog,0);
			enable_obj_w32(hw, j + ID_CHANGE_PAD, (tmp != cfg2type(key_config_tmp[j][NUM_KEY_CFGS-1].analog,tmp)?TRUE:FALSE));
			for (i = 0; i < MAX_PAD_BITS; i++) {
				SendMessage(GetDlgItem(hw,ID_PADTYPE + j * MAX_PAD_BITS + i), BM_SETCHECK, (key_config_tmp[j][NUM_KEY_CFGS-1].analog & 1<<i) ? BST_CHECKED : BST_UNCHECKED, 0);
			}
			refresh_btns(hw, j);
		}
		for (i = 0; threshold_list[i]; i++) {
			HWND h;
			char c[PGN_MAX_FIGURES + 1];
			h = GetDlgItem(hw, ID_CFG_THRESHOLD);
			sprintf(c, "%d", threshold_list[i]);
			SendMessage(h, CB_ADDSTRING, 0, (LPARAM) c);
			if (!i) SetWindowText(h, c);
		}
		if (!init_update_release(hw, 10, &input_state)) {
			pushed = 1;
		} else {
			pushed = 0;
		}
		SendMessage(GetDlgItem(hw,ID_SAVE_INI), BM_SETCHECK, (save_ini_temp) ? BST_CHECKED : BST_UNCHECKED, 0);
		return TRUE;

	case WM_COMMAND:
		switch (LOWORD(w)) {
		case IDOK:
			for (j = 0; j < 2; j++) {
				for (i = 0; i < NUM_KEY_CFGS; i++) {
					key_config[j][i] = key_config_tmp[j][i];
				}
			}
			save_ini = save_ini_temp;
			pgn_save_conf();
		case IDCANCEL:
			release_di();
			EndDialog(hw, TRUE);
			return TRUE;
		case ID_PUSHED:
			enable_obj_w32(hw, ID_PUSHED, FALSE);
			if (!init_update_release(hw, 250, &input_state)) {
				pushed = 1;
			}
			enable_obj_w32(hw, ID_PUSHED, TRUE);
			break;
		case ID_SAVE_INI:
			save_ini_temp = (SendMessage(GetDlgItem(hw, ID_SAVE_INI), BM_GETCHECK, 0, 0) == BST_CHECKED);
			break;
		case ID_LOAD_REG:
		case ID_LOAD_INI:
			for (j = 0; j < 2; j++) {
				read_config(key_config_tmp[j], j, (LOWORD(w) == ID_LOAD_INI), key_config_default[j != 0]);
				for (i = 0; i < NUM_DLG_CFGS; i++) {
					set_btn_text(hw, j * NUM_DLG_CFGS + i + ID_BTN_DOWN, &key_config_tmp[j][i]);
				}
				tmp = cfg2type(key_config_tmp[j][NUM_KEY_CFGS-1].analog, 0);
				enable_obj_w32(hw, j + ID_CHANGE_PAD, (tmp != cfg2type(key_config_tmp[j][NUM_KEY_CFGS-1].analog, tmp) ? TRUE : FALSE));
				for (i = 0; i < MAX_PAD_BITS; i++) {
					SendMessage(GetDlgItem(hw,ID_PADTYPE + j * MAX_PAD_BITS + i), BM_SETCHECK, (key_config_tmp[j][NUM_KEY_CFGS-1].analog & 1<<i) ? BST_CHECKED : BST_UNCHECKED, 0);
				}
				refresh_btns(hw, j);
			}
			break;
		case ID_AUTOFIRE:
			threshold = get_threshold(hw, ID_CFG_THRESHOLD);
			DialogBox(hinst, MAKEINTRESOURCE(IDPGN_AF_DLG) ,hw, (DLGPROC) AutofireDlgProc);
			break;
		case ID_BRAVO:
			threshold = get_threshold(hw, ID_CFG_THRESHOLD);
			DialogBox(hinst, MAKEINTRESOURCE(IDPGN_BRAVO_DLG) ,hw, (DLGPROC) BravoDlgProc);
			break;
		default:
			if (ID_PADTYPE <= w && w < MAX_PAD_BITS * 2 + ID_PADTYPE) {
				if (SendMessage(GetDlgItem(hw, w), BM_GETCHECK, 0, 0) == BST_CHECKED) {
					key_config_tmp[(LOWORD(w)-ID_PADTYPE)/MAX_PAD_BITS][NUM_KEY_CFGS-1].analog |= 1 << ((LOWORD(w)-ID_PADTYPE)%MAX_PAD_BITS);
				} else {
					key_config_tmp[(LOWORD(w)-ID_PADTYPE)/MAX_PAD_BITS][NUM_KEY_CFGS-1].analog &= ~(1 << (LOWORD(w)-ID_PADTYPE)%MAX_PAD_BITS);
				}
				j = (LOWORD(w)-ID_PADTYPE) / MAX_PAD_BITS;
				refresh_btns(hw, j);
			}
			if (ID_BTN_DOWN <= w && w < NUM_DLG_CFGS * 2 + ID_BTN_DOWN) {
				scan_input(hw,w);
				choose_config((w - ID_BTN_DOWN) / NUM_DLG_CFGS, (w - ID_BTN_DOWN) % NUM_DLG_CFGS, key_config_candidate, key_config_tmp, hw, ((w - ID_BTN_DOWN) % NUM_DLG_CFGS) < 8 );
				set_btn_text(hw, w , &key_config_tmp[(w - ID_BTN_DOWN) / NUM_DLG_CFGS][(w - ID_BTN_DOWN) % NUM_DLG_CFGS]);
				enable_obj_w32(hw, w, TRUE);
			}
			if (ID_CHANGE_PAD <= w && w < ID_CHANGE_PAD + 2) {
				scan_input(hw, w);
				choose_config((w - ID_CHANGE_PAD), NUM_KEY_CFGS - 1, key_config_candidate, key_config_tmp, hw, 1 );
				set_btn_text(hw, w , &key_config_tmp[w - ID_CHANGE_PAD][NUM_KEY_CFGS - 1]);
				enable_obj_w32(hw, w, TRUE);
			}
		}
	}
	return FALSE;
}

/*
 * pgn_configure() is called by PADconfigure(), reads temporally config
 * and calls dialog procedure.
 */
void pgn_configure()
{
	DialogBox(hinst, MAKEINTRESOURCE(IDPGN_CFG_DLG), GetActiveWindow(), (DLGPROC)ConfigDlgProc);
}

